"""Utility methods."""
from text_detector import TextBox
from typing import List, Sequence

import cv2
import itertools
import numpy as np


def draw_text_box(tb: TextBox, image: np.ndarray, color=(0, 0, 255), size=2):
    """Draws a red rectangle around the text box. Modifies the array in place."""
    cv2.rectangle(image, (tb.x, tb.y), (tb.x + tb.h, tb.y + tb.w), color, size)


def draw_text_boxes(tbs: Sequence[TextBox], in_path: str, out_path: str, color=(0, 0, 255)):
    """Draws red rectangles around the given text boxes."""
    image = cv2.imread(in_path)
    for tb in tbs:
        draw_text_box(tb, image, color)
    cv2.imwrite(out_path, image)


def intersection_over_union(box1: TextBox, box2: TextBox):
    # Determine the (x, y)-coordinates of the intersection rectangle.
    xa = max(box1.x, box2.x)
    ya = max(box1.y, box2.y)
    xb = min(box1.x + box1.h, box2.x + box2.h)
    yb = min(box1.y + box1.w, box2.y + box2.w)
    # Compute the area of intersection rectangle.
    intersection_area = max(0, xb - xa + 1) * max(0, yb - ya + 1)
    # Compute the area of both the prediction and ground-truth rectangles
    box1_area = (box1.h + 1) * (box1.w + 1)
    box2_area = (box2.h + 1) * (box2.w + 1)
    iou = intersection_area / float(box1_area + box2_area - intersection_area)
    return iou


def multi_intersection_over_union(detected_boxes: Sequence[TextBox], gold_boxes: Sequence[TextBox]):
    """Computes average IOU across the detected boxes.

    For a particular detected box, finds the golden box with maximum IOU. Any gold box that doesn't intersect with any
    of the detected boxes contributes to the average with a 0, to penalize poor recall.

    This might not necessarily be the standard in academia, but we just need a consistent way of evaluating the quality
    of various text detectors in our pipeline.
    """
    matched_golden_boxes = set()
    max_ious = []
    for db in detected_boxes:
        ious = [intersection_over_union(db, gb) for gb in gold_boxes]
        max_iou = max(ious)
        max_ious.append(max_iou)
        for idx, iou in enumerate(ious):
            if iou == max_iou:
                matched_golden_boxes.add(idx)

    # For every golden box that did not match against a detected box, add a 0.
    for idx in range(len(gold_boxes)):
        if idx not in matched_golden_boxes:
            max_ious.append(0.0)

    return np.mean(max_ious)


def merge_nearby_boxes(boxes: Sequence[TextBox], max_distance) -> Sequence[TextBox]:
    """Merges boxes that are less than `max_distance` pixels apart on both the x and y axes."""
    if len(boxes) <= 1:
        return boxes

    def overlap_x(box1: TextBox, box2: TextBox) -> int:
        return min(box1.x + box1.h, box2.x + box2.h) - max(box1.x, box2.x)

    def overlap_y(box1: TextBox, box2: TextBox) -> int:
        return min(box1.y + box1.w, box2.y + box2.w) - max(box1.y, box2.y)

    def should_merge(box1: TextBox, box2: TextBox) -> bool:
        # Boxes need to overlap on one axis and be close to each other on the other axis.
        # Note that the inverse of overlap is distance.
        x_overlap = overlap_x(box1, box2)
        y_overlap = overlap_y(box1, box2)
        return (x_overlap > 0 and -y_overlap < max_distance) or (y_overlap > 0 and -x_overlap < max_distance)

    def merge(bs: Sequence[TextBox]) -> TextBox:
        """Merges boxes into the smallest enclosing rectangle."""
        tl = (min([b.x for b in bs]), min([b.y for b in bs]))
        br = (max([b.x + b.h for b in bs]), max([b.y + b.w for b in bs]))
        return TextBox(x=tl[0], y=tl[1], h=br[0] - tl[0], w=br[1] - tl[1])

    def merge_with_box(ref_box: TextBox, other_boxes: List[TextBox]) -> List[TextBox]:
        """Merges `ref_box` with boxes from `other_boxes` that are close enough. Returns the other boxes unchanged."""
        should_merge_with_ref = [should_merge(ref_box, box) for box in other_boxes]

        if sum(should_merge_with_ref) == 0:
            return [ref_box] + other_boxes

        to_merge = list(itertools.compress(other_boxes, should_merge_with_ref))
        merged_box = merge([ref_box] + to_merge)
        should_keep = [not should for should in should_merge_with_ref]
        to_keep = list(itertools.compress(other_boxes, should_keep))
        return [merged_box] + to_keep

    curr_boxes = boxes
    ref_idx = 0
    while ref_idx < len(curr_boxes) - 1:
        before = curr_boxes[:ref_idx]
        after = merge_with_box(curr_boxes[ref_idx], curr_boxes[ref_idx:])
        if len(before) + len(after) == len(curr_boxes):
            # No merge happened. We can advance the index.
            ref_idx += 1
        curr_boxes = before + after

    return curr_boxes



"""Runs text detection over the folders in an image. Draws red rectangles around the detected text.

TODO: Remove this file from the library before publishing.
"""
from absl import app
from absl import logging
from absl import flags
from detextify.text_detector import AzureTextDetector, PaddleTextDetector

import annotation_parser
import cv2
import json
import glob
import numpy as np
import os
import utils

flags.DEFINE_string("input_dir", None, "Directory with input images.")
flags.DEFINE_string("output_dir", None, "Directory with output images, with text boxes drawn on the input image.")
flags.DEFINE_enum("mode", "PADDLE", ["AZURE", "PADDLE"], "Type of text detection.")
flags.DEFINE_string("azure_endpoint", None,
                    "Needed when --mode=AZURE. Should have the form https://<name>.cognitiveservices.azure.com/")
flags.DEFINE_string("azure_key", None, "Needed when --mode=AZURE.")
flags.DEFINE_string("annotations_file", None,
                    "Optional path to file with ground labels. If specified, we calculate IOU against text detection.")
flags.mark_flags_as_required(["input_dir", "output_dir"])
FLAGS = flags.FLAGS


def main(_):
    if FLAGS.mode == "PADDLE":
        detector = PaddleTextDetector()
    elif FLAGS.mode == "AZURE":
        if not FLAGS.azure_endpoint:
            raise ValueError("Please specify --azure_endpoint")
        if not FLAGS.azure_key:
            raise ValueError("Please specify --azure_key")
        detector = AzureTextDetector(FLAGS.azure_endpoint, FLAGS.azure_key)
    else:
        raise NotImplemented(f"Mode `{FLAGS.mode}` for text detection is not implemented.")

    if not os.path.exists(FLAGS.output_dir):
        os.makedirs(FLAGS.output_dir)

    if not FLAGS.annotations_file:
        # Simply draw the detected text boxes.
        for image_path in glob.glob(os.path.join(FLAGS.input_dir, "*")):
            text_boxes = detector.detect_text(image_path)
            name, ext = os.path.splitext(os.path.basename(image_path))
            output_path = os.path.join(FLAGS.output_dir, name + f"_boxes_{FLAGS.mode}" + ext)
            utils.draw_text_boxes(text_boxes, image_path, output_path)
        return

    # Evaluate detected text boxes against gold annotations.
    ious = []
    with open(os.path.join(FLAGS.output_dir, "iou.txt"), "w") as fout:
        with open(FLAGS.annotations_file) as f:
            annotations = json.load(f)
            for annotation in annotations:
                image_basename = annotation_parser.get_image_basename(annotation)
                image_path = os.path.join(FLAGS.input_dir, image_basename)
                if not os.path.exists(image_path):
                    logging.error(f"Unable to find image {image_path}.")
                    continue

                detected_boxes = detector.detect_text(image_path)
                # Text detection might produce granular boxes (e.g. one per word); merge them to resemble the annotations.
                merged_detected_boxes = utils.merge_nearby_boxes(detected_boxes, max_distance=30)
                golden_boxes = annotation_parser.convert_to_text_boxes(annotation)

                # For debugging purposes, draw all these boxes.
                image = cv2.imread(image_path)  # (B, G, R)
                for box in detected_boxes:
                    utils.draw_text_box(box, image, color=(0, 255, 0), size=5)
                for box in merged_detected_boxes:
                    utils.draw_text_box(box, image, color=(255, 0, 0), size=2)
                for box in golden_boxes:
                    utils.draw_text_box(box, image, color=(0, 0, 255), size=2)
                output_path = os.path.join(FLAGS.output_dir, f"final_merged_{image_basename}")
                cv2.imwrite(output_path, image)

                iou = utils.multi_intersection_over_union(merged_detected_boxes, golden_boxes)
                print(f"IOU = {iou} for {image_basename}")
                fout.write(f"IOU = {iou} for {image_basename}\n")
                ious.append(iou)

        print("Average IOU across %d images: %f" % (len(ious), np.mean(ious)))
        fout.write("Average IOU across %d images: %f" % (len(ious), np.mean(ious)))


if __name__ == "__main__":
    app.run(main)



"""Interfaces for text detection."""
from absl import logging
from dataclasses import dataclass
from typing import Sequence

import time

from azure.cognitiveservices.vision.computervision import ComputerVisionClient
from azure.cognitiveservices.vision.computervision.models import OperationStatusCodes
from msrest.authentication import CognitiveServicesCredentials
from paddleocr import PaddleOCR


@dataclass
class TextBox:
  # (x, y) is the top left corner of a rectangle; the origin of the coordinate system is the top-left of the image.
  # x denotes the vertical axis, y denotes the horizontal axis (to match the traditional indexing in a matrix).
  x: int
  y: int
  h: int
  w: int
  text: str = None


class TextDetector:
  def detect_text(self, image_filename: str) -> Sequence[TextBox]:
    pass


class AzureTextDetector(TextDetector):
  """Calls the Computer Vision endpoint from Microsoft Azure. Promises to work with images in the wild."""

  def __init__(self, endpoint, key):
    self.client = ComputerVisionClient(endpoint, CognitiveServicesCredentials(key))

  def detect_text(self, image_filename: str) -> Sequence[TextBox]:
    read_response = self.client.read_in_stream(open(image_filename, "rb"), raw=True)

    # Get the operation location (URL with an ID at the end) from the response
    read_operation_location = read_response.headers["Operation-Location"]
    # Grab the ID from the URL
    operation_id = read_operation_location.split("/")[-1]

    # Call the "GET" API and wait for it to retrieve the results
    while True:
      read_result = self.client.get_read_result(operation_id)
      if read_result.status not in ['notStarted', 'running']:
        break
      time.sleep(1)

    text_boxes = []
    if read_result.status == OperationStatusCodes.succeeded:
      for text_result in read_result.analyze_result.read_results:
        for line in text_result.lines:
          # line.bounding_box contains the 4 corners of a polygon (not necessarily a rectangle).
          # To keep things simple, we turn them into rectangles. There are two ways: (1) use the rectangle
          # defined by the top-left and bottom-right corners, or (2) use the rectangle that encompasses the
          # entire polygon. (1) will lead to smaller surfaces, (2) to bigger surfaces.

          # Implementation for (1)
          # tl_x, tl_y = line.bounding_box[0:2]   # top left
          # br_x, br_y = line.bounding_box[4:6]   # bottom right
          # w = br_x - tl_x
          # h = br_y - tl_y

          # Implementation for (2)
          xs = [point for idx, point in enumerate(line.bounding_box) if idx % 2 == 0]
          ys = [point for idx, point in enumerate(line.bounding_box) if idx % 2 == 1]
          tl_x = min(xs)
          tl_y = min(ys)
          h = max(xs) - tl_x
          w = max(ys) - tl_y

          if h < 0 or w < 0:
            logging.error(f"Malformed bounding box from Azure: {line.bounding_box}")

          text_boxes.append(TextBox(int(tl_x), int(tl_y), int(h), int(w), line.text))
    return text_boxes



class PaddleTextDetector(TextDetector):
  """Uses PaddleOCR for text detection: https://github.com/PaddlePaddle/PaddleOCR"""

  def __init__(self):
    self.ocr = PaddleOCR(use_angle_cls=True, use_gpu=True, lang="en")

  def detect_text(self, image_path: str) -> Sequence[TextBox]:
    result = self.ocr.ocr(image_path, cls=True)[0]
    text_boxes = []
    for line in result:
      points = line[0]
      text = line[1][0]
      # These points are not necessarily a rectangle, but rather a polygon.
      # We'll find the smallest enclosing rectangle.
      xs = [point[0] for point in points]
      ys = [point[1] for point in points]
      tl_x = min(xs)
      tl_y = min(ys)
      h = max(xs) - tl_x
      w = max(ys) - tl_y

      if h < 0 or w < 0:
        logging.error(f"Malformed bounding box from Paddle: {points}")

      text_boxes.append(TextBox(int(tl_x), int(tl_y), int(h), int(w), text))
    return text_boxes



"""In-painting models."""
import io
import openai
import replicate
import requests
import tempfile
import torch

from PIL import Image, ImageDraw
from diffusers import StableDiffusionInpaintPipeline
from text_detector import TextBox
from typing import Sequence


class Inpainter:
  """Interface for in-painting models."""
  # TODO(julia): Run some experiments to determine the best prompt.
  PROMPT = "plain background"

  def inpaint(self, in_image_path: str, text_boxes: Sequence[TextBox], out_image_path: str):
    pass


class DalleInpainter(Inpainter):
  """In-painting model that calls the DALL-E API."""

  def __init__(self, openai_key: str):
    openai.api_key = openai_key

  @staticmethod
  def _make_mask(text_boxes: Sequence[TextBox], height: int, width: int) -> bytes:
    """Returns an .png where the text boxes are transparent."""
    mask = Image.new("RGBA", (width, height), (0, 0, 0, 1))  # fully opaque
    mask_draw = ImageDraw.Draw(mask)
    for text_box in text_boxes:
      mask_draw.rectangle(xy=(text_box.x, text_box.y, text_box.x + text_box.h, text_box.y + text_box.w),
                          fill=(0, 0, 0, 0))  # fully transparent
    # Convert mask to bytes.
    bytes_arr = io.BytesIO()
    mask.save(bytes_arr, format="PNG")
    return bytes_arr.getvalue()

  def inpaint(self, in_image_path: str, text_boxes: Sequence[TextBox], out_image_path: str):
    image = Image.open(in_image_path)  # open the image to inspect its size

    response = openai.Image.create_edit(
        image=open(in_image_path, "rb"),
        mask=self._make_mask(text_boxes, image.height, image.width),
        prompt=Inpainter.PROMPT,
        n=1,
        size=f"{image.height}x{image.width}"
    )
    url = response['data'][0]['url']
    out_image_data = requests.get(url).content
    out_image = Image.open(io.BytesIO(out_image_data))
    out_image.save(out_image_path)


def make_black_and_white_mask(text_boxes: Sequence[TextBox], height: int, width: int, out_mask_path: str):
  """Returns a black image with white rectangles where the text boxes are."""
  mask = Image.new("RGB", (width, height), (0, 0, 0))  # black
  mask_draw = ImageDraw.Draw(mask)
  for text_box in text_boxes:
    mask_draw.rectangle(xy=(text_box.x, text_box.y, text_box.x + text_box.h, text_box.y + text_box.w),
                        fill=(255, 255, 255))  # white
  mask.save(out_mask_path)


class ReplicateInpainter(Inpainter):
  SD_INPAINTING_V2 = "cjwbw/stable-diffusion-v2-inpainting"
  SD_INPAINTING_V2_VERSION = "f9bb0632bfdceb83196e85521b9b55895f8ff3d1d3b487fd1973210c0eb30bec"

  def __init__(self, replicate_token: str, model_name=SD_INPAINTING_V2, model_version=SD_INPAINTING_V2_VERSION):
    replicate_client = replicate.Client(api_token=replicate_token)
    self.model = replicate_client.models.get(model_name).versions.get(model_version)

  def inpaint(self, in_image_path: str, text_boxes: Sequence[TextBox], out_image_path: str):
    image = Image.open(in_image_path)  # open the image to inspect its size
    mask_temp_file = tempfile.NamedTemporaryFile(suffix=".jpeg")
    make_black_and_white_mask(text_boxes, image.height, image.width, mask_temp_file.name)

    url = self.model.predict(prompt=Inpainter.PROMPT,
                             image=open(in_image_path, "rb"),
                             mask=open(mask_temp_file.name, "rb"),
                             num_outputs=1)[0]

    out_image_data = requests.get(url).content
    out_image = Image.open(io.BytesIO(out_image_data))
    out_image.save(out_image_path)



class DiffusersSDInpainter(Inpainter):
  """Uses a Stable Diffusion model from HuggingFace for in-painting."""

  def __init__(self, pipe: StableDiffusionInpaintPipeline = None):
    if pipe is None:
      if not torch.cuda.is_available():
        raise Exception("You need a GPU + CUDA to run this model locally.")

      self.pipe = StableDiffusionInpaintPipeline.from_pretrained(
          "stabilityai/stable-diffusion-2-inpainting",
          revision="fp16",
          torch_dtype=torch.float16).to("cuda")
    else:
      self.pipe = pipe

  def inpaint(self, in_image_path: str, text_boxes: Sequence[TextBox], out_image_path: str):
    image = Image.open(in_image_path)

    mask_temp_file = tempfile.NamedTemporaryFile(suffix=".jpeg")
    make_black_and_white_mask(text_boxes, image.height, image.width, mask_temp_file.name)
    mask_image = Image.open(mask_temp_file.name)

    out_image = self.pipe(prompt=Inpainter.PROMPT, image=image, mask_image=mask_image).images[0]
    out_image.save(out_image_path)



import os
from typing import Dict, Optional, Sequence
from text_detector import TextBox


def convert_to_text_box(result: Dict) -> Optional[TextBox]:
    if 'original_width' not in result or 'original_height' not in result:
        return None
    
    value = result['value']
    # Since we made the annotations, we flipped the meaning of "height" and "weight" in TextBox to be more intuitive.
    # Now, height is for vertical and width is for horizontal.
    original_h = result["original_width"]   # notice meaning flip
    original_w = result["original_height"]  # notice meaning flip

    if all([key in value for key in ['x', 'y', 'height', 'width']]):
        x = value["x"]
        y = value["y"]
        h = value["width"]   # notice meaning flip
        w = value["height"]  # notice meaning flip

        return TextBox(x=int(x * original_h / 100.0),
                       y=int(y * original_w / 100.0),
                       h=int(h * original_h / 100.0),
                       w=int(w * original_w / 100.0))


def convert_to_text_boxes(annotation: Dict) -> Sequence[TextBox]:
    assert len(annotation["annotations"]) == 1
    gold = annotation["annotations"][0]
    text_boxes = [convert_to_text_box(result) for result in gold["result"]]
    return [tb for tb in text_boxes if tb is not None]


def get_image_basename(annotation: Dict) -> str:
    return os.path.basename("".join(annotation["data"]["data"].split("-")[1:]))



from absl import logging
from inpainter import Inpainter
from text_detector import TextDetector


class Detextifier:
    def __init__(self, text_detector: TextDetector, inpainter: Inpainter):
        self.text_detector = text_detector
        self.inpainter = inpainter

    def detextify(self, in_image_path: str, out_image_path: str, max_retries=5):
        to_inpaint_path = in_image_path
        for i in range(max_retries):
            logging.info(f"Inpainting iteration {i} for image {in_image_path}...")
            text_boxes = self.text_detector.detect_text(to_inpaint_path)
            logging.info(f"Detected {len(text_boxes)} text boxes.")
            if not text_boxes:
                break
            self.inpainter.inpaint(to_inpaint_path, text_boxes, out_image_path)
            to_inpaint_path = out_image_path



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['detextify']

package_data = \
{'': ['*'], 'detextify': ['config/*']}

install_requires = \
['absl-py>=1.3.0,<2.0.0',
 'accelerate>=0.15.0,<0.16.0',
 'azure-cognitiveservices-vision-computervision>=0.9.0,<0.10.0',
 'diffusers>=0.10.2,<0.11.0',
 'numpy>=1.23.5,<2.0.0',
 'openai>=0.25.0,<0.26.0',
 'opencv-python>=4.6.0.66,<5.0.0.0',
 'paddleocr>=2.6.1.1,<3.0.0.0',
 'paddlepaddle>=2.4.1,<3.0.0',
 'pillow>=9.3.0,<10.0.0',
 'replicate>=0.4.0,<0.5.0',
 'requests>=2.28.1,<3.0.0',
 'torch>=1.13.0,<2.0.0',
 'transformers>=4.25.1,<5.0.0']

setup_kwargs = {
    'name': 'detextify',
    'version': '0.1.0',
    'description': '',
    'long_description': 'Library to remove unwanted pseudo-text from images generated by AI\n===\nTo run the main script (which will eventually be removed from the library):\npoetry run python detextify/main.py\n',
    'author': 'Julia',
    'author_email': 'turc.raluca@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
